import codecs
import json
try:
    from urllib.parse import urljoin
    from urllib.parse import urlencode
except ImportError:
    from urlparse import urljoin
    from urllib import urlencode

from django.conf import settings
from django.conf.urls import patterns
from django.conf.urls import url
from django.core import signing
from django.core.urlresolvers import reverse

class InvalidNextUrl(Exception):
    pass

class Next():
    salt = 'facebook_auth.urls.Next'

    def encode(self, data):
        data = signing.dumps(data, salt=self.salt)
        return urlencode({'next': data})

    def decode(self, data):
        try:
            return signing.loads(data, salt=self.salt)
        except signing.BadSignature:
            raise InvalidNextUrl()


def redirect_uri(next, close):
    return urljoin(
        settings.FACEBOOK_CANVAS_URL,
        reverse('facebook-auth-handler') + "?" +
        Next().encode({'next': next, 'close': close})
    )

urlpatterns = patterns('facebook_auth.views',
    url(r'^handler$', 'handler', name='facebook-auth-handler'),
)



import importlib
import logging

import facepy
from facepy.exceptions import FacebookError

from django.conf import settings
from django.utils import timezone


def get_class(class_name):
    module_name, class_name = class_name.rsplit(".", 1)

    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def get_graph_observer_classes(graph_observers):
    return list(map(get_class, graph_observers))


FACEBOOK_GRAPH_OBSERVERS = getattr(settings, 'FACEBOOK_GRAPH_OBSERVERS', [])
GRAPH_OBSERVER_CLASSES = get_graph_observer_classes(FACEBOOK_GRAPH_OBSERVERS)


logger = logging.getLogger(__name__)


class ObservableGraphAPI(facepy.GraphAPI):
    def __init__(self, *args, **kwargs):
        super(ObservableGraphAPI, self).__init__(*args, **kwargs)
        self.session = ObservableSession(self.session)

    def _query(self, *args, **kwargs):
        handlers = FacebookConnectionObservers()
        self.session.observers.append(handlers)
        try:
            response = super(ObservableGraphAPI, self)._query(*args, **kwargs)
        except FacebookError as e:
            handlers.handle_error(e)
            raise
        finally:
            self.session.observers.remove(handlers)
            handlers.finalize()
        return response

    def batch(self, requests):
        logger.warn('Errors of batch method are not watched')
        return super(ObservableGraphAPI, self).batch(requests)


class ObservableSession(object):
    def __init__(self, other_session):
        self.other_session = other_session
        self.observers = []

    def request(self, *args, **kwargs):
        self.notify_request(*args, **kwargs)
        response = self.other_session.request(*args, **kwargs)
        self.notify_response(response)
        return response

    def notify_request(self, *args, **kwargs):
        for observer in self.observers:
            observer.handle_request(*args, **kwargs)

    def notify_response(self, response):
        for observer in self.observers:
            observer.handle_response(response)


class FacebookConnectionObservers(object):
    def __init__(self):
        self.request = None
        self.response = None
        self.error = None
        self.start = timezone.now()

    def handle_request(self, *args, **kwargs):
        self.request = RequestInfo(*args, **kwargs)

    def handle_response(self, response):
        self.response = response

    def handle_error(self, error):
        self.error = error

    def finalize(self):
        time = timezone.now() - self.start
        for observer_class in GRAPH_OBSERVER_CLASSES:
            observer = observer_class(
                self.request, self.response, self.error, time)
            observer.handle_facebook_communication()


class RequestInfo(object):
    def __init__(self, method, url, **kwargs):
        self.url = url
        self.method = method
        self.kwargs = kwargs



try:
    from unittest import mock
    from urllib import parse
    MOCK_CLASS_NAME = 'unittest.mock.Mock'
except ImportError:
    import mock
    import urlparse as parse
    MOCK_CLASS_NAME = 'mock.Mock'

import datetime

from django import test

from facepy.exceptions import FacebookError
import pytz

from facebook_auth.backends import _truncate as truncate
from facebook_auth.backends import UserFactory
from facebook_auth import forms
from facebook_auth import graph_api
from facebook_auth import models
from facebook_auth import urls
from facebook_auth import views


class TruncaterTest(test.SimpleTestCase):
    def test_empty(self):
        self.assertEqual('', truncate('', 30))

    def test_no_cutting(self):
        word = 'abcde'
        self.assertEqual(word, truncate(word, 10))
        self.assertEqual(word, truncate(word, 6))
        self.assertEqual(word, truncate(word, 5))

    def test_cutting(self):
        word = 'abcde'
        self.assertEqual(word[:4], truncate(word, 4))
        self.assertEqual(word[:3], truncate(word, 3))
        self.assertEqual('', truncate(word, 0))

    def test_to_zero_cut(self):
        word = 'abcde'
        self.assertEqual(word, truncate(word, 10, to_zero=True))
        self.assertEqual(word, truncate(word, 6, to_zero=True))
        self.assertEqual(word, truncate(word, 5, to_zero=True))
        self.assertEqual('', truncate(word, 4, to_zero=True))
        self.assertEqual('', truncate(word, 3, to_zero=True))
        self.assertEqual('', truncate(word, 0, to_zero=True))


class UserFactoryTest(test.TestCase):
    def test_empty(self):
        profile = {
            'id': '1',
            'first_name': '',
            'last_name': '',
            'email': ''
        }
        UserFactory()._product_user('', profile).save()

    def test_no_email(self):
        profile = {
            'id': '1',
            'first_name': '',
            'last_name': '',
        }
        UserFactory()._product_user('', profile).save()

    def test_to_long(self):
        profile = {
            'id': '1',
            'first_name': 'a' * 1000,
            'last_name': 'a' * 1000,
            'email': 'a' * 1000
        }
        user = UserFactory()._product_user('', profile)
        user.save()

        def get_length(field):
            return user._meta.get_field_by_name(field)[0].max_length

        self.assertEqual(user.first_name, 'a' * get_length('first_name'))
        self.assertEqual(user.last_name, 'a' * get_length('last_name'))
        self.assertEqual(user.email, '')


class UserFactoryOnErrorTest(test.TestCase):
    def test_success(self):
        factory = UserFactory()
        factory.graph_api_class = mock.Mock()
        factory.graph_api_class.return_value.get.return_value = {'id': '123'}
        user = factory.get_user("123")
        self.assertEqual(123, user.user_id)

    def test_success_in_retry(self):
        factory = UserFactory()
        factory.graph_api_class = mock.Mock()
        factory.graph_api_class.return_value.get.side_effect = [
            FacebookError("msg", 1),
            FacebookError("msg", 1),
            {'id': '123'}]

        user = factory.get_user("123")
        self.assertEqual(123, user.user_id)

    def test_failure(self):
        factory = UserFactory()
        factory.graph_api_class = mock.Mock()
        factory.graph_api_class.return_value.get.side_effect = [
            FacebookError("msg", 1),
            FacebookError("msg", 1),
            FacebookError("msg", 1),
            {'id': '123'}]
        with self.assertRaises(FacebookError):
            factory.get_user("123")


@mock.patch('django.utils.timezone.now')
@mock.patch('facepy.GraphAPI._query')
@mock.patch('facebook_auth.graph_api.GRAPH_OBSERVER_CLASSES')
class ObservableGraphApiTest(test.SimpleTestCase):
    def test_query_failure(self, observers, query, now):
        now.side_effect = [
            datetime.datetime(year=1, month=1, day=1, minute=1),
            datetime.datetime(year=1, month=1, day=1, minute=2),
        ]
        query.side_effect = FacebookError("msg", 1)
        observer_cls = mock.Mock()
        observers.__iter__.return_value = [observer_cls]
        with self.assertRaises(FacebookError):
            graph_api.ObservableGraphAPI().get('me')
        observer_cls.return_value.handle_facebook_communication.assert_called_once_with()
        observer_cls.assert_called_once_with(None, None, query.side_effect,
                                             datetime.timedelta(minutes=1))

    def test_query_success_string(self, observers, query, now):
        now.side_effect = [
            datetime.datetime(year=1, month=1, day=1, minute=1),
            datetime.datetime(year=1, month=1, day=1, minute=2),
        ]
        query.return_value = 'some string response'
        observer_cls = mock.Mock()
        observers.__iter__.return_value = [observer_cls]
        graph_api.ObservableGraphAPI().get('me')
        observer_cls.return_value.handle_facebook_communication.assert_called_once_with()
        observer_cls.assert_called_once_with(None, None, None,
                                             datetime.timedelta(minutes=1))


class GraphObserversTest(test.SimpleTestCase):
    def test_getting_observer_classes(self):
        classes = graph_api.get_graph_observer_classes([MOCK_CLASS_NAME])
        self.assertEqual([mock.Mock], list(classes))

    def test_iterating_observer_classes_twice(self):
        classes = graph_api.get_graph_observer_classes([MOCK_CLASS_NAME])
        list(classes)
        self.assertEqual([mock.Mock], list(classes))


class UserTokenManagerTest(test.TestCase):
    def test_simple_insert(self):
        manager = models.UserTokenManager
        manager.insert_token('123', 'abc123',
                             datetime.datetime(1989, 2, 25, tzinfo=pytz.utc))
        token = manager.get_access_token('123')
        self.assertEqual('abc123', token.token)

    def test_multiple_inserts(self):
        manager = models.UserTokenManager
        manager.insert_token('123', 'abc123',
                             datetime.datetime(1989, 2, 25, tzinfo=pytz.utc))
        manager.insert_token('456', 'abc456',
                             datetime.datetime(1989, 2, 25, tzinfo=pytz.utc))
        token = manager.get_access_token('123')
        self.assertEqual('abc123', token.token)

        token2 = manager.get_access_token('456')
        self.assertEqual('abc456', token2.token)

    def test_invalidating_token(self):
        manager = models.UserTokenManager
        manager.insert_token('123', 'abc123',
                             datetime.datetime(1989, 2, 25, tzinfo=pytz.utc))
        manager.invalidate_access_token('abc123')
        self.assertRaises(models.UserToken.DoesNotExist,
                          manager.get_access_token, '123')


class TestParseFacebookResponse(test.SimpleTestCase):
    def test_without_data(self):
        response = forms.parse_facebook_response({}, '123')
        self.assertEqual(response.is_valid, False)

    def test_with_empty_data(self):
        response = forms.parse_facebook_response({'data': {}}, '123')
        self.assertEqual(response.is_valid, False)

    def test_if_original_dict_is_not_modified(self):
        data = {}
        input_json = {'data': data}
        forms.parse_facebook_response(input_json, '123')
        self.assertEqual({}, data)
        self.assertEqual({'data': {}}, input_json)

    def test_is_valid_as_string(self):
        data = {
            'expires_at': 12341234,
            'is_valid': 'foo',
            'scopes[]': 'foo,bar',
            'user_id': '123',
        }
        response = forms.parse_facebook_response({'data': data}, '123')
        self.assertEqual(response.is_valid, True)

    def test_valid_response(self):
        data = {
            'expires_at': 12341234,
            'is_valid': True,
            'scopes[]': 'foo,bar',
            'user_id': '123',
        }
        response = forms.parse_facebook_response({'data': data}, '123')
        self.assertEqual(response.is_valid, True)

    def test_valid_real_data(self):
        data = {
            "expires_at": 1403429380,
            "scopes": [
              "public_profile",
              "basic_info",
              "publish_checkins",
              "status_update",
              "photo_upload",
              "video_upload",
              "email",
              "create_note",
              "share_item",
              "publish_stream",
              "publish_actions",
              "user_friends"
            ],
            "app_id": 423260947733647,
            "application": "'Social WiFi'",
            "issued_at": 1398245380,
            "is_valid": True,
            "user_id": 1000066666,
        }
        response = forms.parse_facebook_response({'data': data}, '123')
        self.assertEqual(response.is_valid, True)

    def test_test_strange_types(self):
        data = {
            'expires_at': {},
            'is_valid': [],
            'scopes[]': {},
            'user_id': 1.1,
        }
        response = forms.parse_facebook_response({'data': data}, '123')
        self.assertEqual(response.is_valid, False)

    def test_data_as_list(self):
        response = forms.parse_facebook_response({'data': []}, '123')
        self.assertEqual(response.is_valid, False)

    def test_data_as_int(self):
        response = forms.parse_facebook_response({'data': []}, '123')
        self.assertEqual(response.is_valid, False)

    def test_bool_response(self):
        response = forms.parse_facebook_response(False, '123')
        self.assertEqual(response.is_valid, False)


class TestDebugAllTokensForUser(test.TestCase):
    @mock.patch.object(models, 'FacebookTokenManager')
    def test_positive_scenario(self, FacebookTokenManager):
        manager = FacebookTokenManager.return_value
        parsed_data = manager.debug_token.return_value
        parsed_data.token = 'token1212'
        parsed_data.expires = datetime.datetime(2014, 2, 2, tzinfo=pytz.utc)
        token_manager = models.UserTokenManager()
        token_manager.insert_token('123', 'token1212', "2014-02-02")
        models.debug_all_tokens_for_user('123')
        token = token_manager.get_access_token('123')
        self.assertFalse(token.deleted)

    @mock.patch.object(models, 'FacebookTokenManager')
    def test_negative_scenario(self, FacebookTokenManager):
        manager = FacebookTokenManager.return_value
        manager.debug_token.side_effect = ValueError
        token_manager = models.UserTokenManager()
        token_manager.insert_token('123', 'token1212', "2014-02-02")
        models.debug_all_tokens_for_user('123')
        self.assertRaises(models.UserToken.DoesNotExist,
                          token_manager.get_access_token, '123')


class TestNextUrl(test.TestCase):
    def test_invalid_next(self):
        with self.assertRaises(urls.InvalidNextUrl):
            urls.Next().decode('1:2:3')

    def test_invalid_next_format(self):
        with self.assertRaises(urls.InvalidNextUrl):
            urls.Next().decode('this is not valid signature')

    def test_empty_next(self):
        with self.assertRaises(urls.InvalidNextUrl):
            urls.Next().decode('')


class HandlerAcceptanceTest(test.TestCase):
    @mock.patch('facebook_auth.views.authenticate')
    def test_valid_next(self, authenticate):
        authenticate.return_value = None
        encoded_next = urls.Next().encode({
            'next': 'http://next.example.com',
            'close': 'http://close.example.com'})
        next_value = parse.parse_qs(encoded_next)['next'][0]
        request = mock.Mock(GET={'next': next_value, 'code': 'code'},
                            method='GET')
        with self.settings(FACEBOOK_CANVAS_URL='http://example.com'):
            response = views.handler(request)
        self.assertEqual(302, response.status_code)
        self.assertEqual('http://next.example.com', response['Location'])

    def test_invalid_next(self):
        request = mock.Mock(GET={'next': 'a:b:c', 'code': 'code'},
                            method='GET')
        response = views.handler(request)
        self.assertEqual(400, response.status_code)



import logging

from django.contrib.auth import authenticate
from django.contrib.auth import login
from django import http
from django.views import generic

import facepy

from facebook_auth import urls


logger = logging.getLogger(__name__)


class Handler(generic.View):
    def get(self, request):
        try:
            self.next_url = urls.Next().decode(request.GET['next'])
        except urls.InvalidNextUrl:
            logger.warning('Invalid facebook handler next.',
                           extra={'request': request})
            return http.HttpResponseBadRequest()
        if 'code' in request.GET:
            try:
                self.login()
            except facepy.FacepyError as e:
                return self.handle_facebook_error(e)
            response = http.HttpResponseRedirect(self.next_url['next'])
            response["P3P"] = ('CP="IDC DSP COR ADM DEVi TAIi PSA PSD IVAi'
                               ' IVDi CONi HIS OUR IND CNT"')
        else:
            response = http.HttpResponseRedirect(self.next_url['close'])
        return response

    def login(self):
        user = authenticate(
            code=self.request.GET['code'],
            redirect_uri=urls.redirect_uri(self.next_url['next'],
                                           self.next_url['close']))
        if user:
            login(self.request, user)

    def handle_facebook_error(self, e):
        return http.HttpResponseRedirect(self.next_url['next'])

handler = Handler.as_view()
